#ifndef DF_OBJECT_H
#define DF_OBJECT_H

#include <glib.h>

#define DF__OBJECT_DECLARATIONS(name,pname)									\
	typedef union _ ## name ## __Cast name ## __Cast;					\
	typedef struct _ ## name ## __Methods name ## __Methods;				\
	typedef struct _ ## name ## __Interfaces name ## __Interfaces;				\
	typedef struct _ ## name name;								\
												\
	union _ ## name ## __Cast {								\
		name ## __CAST									\
	};											\
	struct _ ## name ## __Methods {								\
		name ## __METHODS								\
	};											\
	struct _ ## name ## __Interfaces {							\
		name ## __INTERFACES								\
	};											\
												\
	struct _ ## name {									\
		guint type;									\
		name ## __Cast       cast;							\
		name ## __Methods   *_m;							\
		name ## __Interfaces *_i;							\
		name ## __DATA									\
	};											\
												\
	extern pname ## __Methods * name ## __ParentMethods;					\
	extern pname ## __Interfaces * name ## __ParentInterfaces;				\
												\
	guint name ## __get_type(void);

#define DF__OBJECT_NEW(name)									\
	((name *) DF__Object__create(name ## __get_type()))

#define DF__OBJECT_DEFINITIONS(name,pname,init_methods,init_interfaces,init_data)		\
	pname ## __Methods * name ## __ParentMethods = NULL;					\
	pname ## __Interfaces * name ## __ParentInterfaces = NULL;				\
	guint name ## __get_type (void) {							\
		static guint type = -1;								\
		if (type == -1) {								\
			guint parent_type;							\
			void (*init_m)(name ## __Methods *m) = init_methods;			\
			void (*init_i)(name ## __Interfaces *i) = init_interfaces;		\
			void (*init_d)(name *self) = init_data;					\
			g_assert(init_m != NULL);						\
			g_assert(init_i != NULL);						\
			g_assert(init_d != NULL);						\
			/* This should be optimized away by the compiler */			\
			if(DF__Object__get_type == name ## __get_type)				\
				parent_type = 0;						\
			else									\
		        	parent_type = pname ## __get_type();				\
			type = DF__Object__init_type(#name,					\
						     parent_type,				\
						     (DF__Object__init)init_m,			\
						     (DF__Object__init)init_i,			\
						     (DF__Object__init_data)init_d,		\
						     sizeof(name),				\
						     sizeof(name ## __Methods),			\
						     sizeof(name ## __Interfaces));		\
			name ## __ParentMethods = DF__OBJECT_METHODS(pname);			\
			name ## __ParentInterfaces = DF__OBJECT_INTERFACES(pname);		\
		}										\
		return type;									\
	}

#define DF__OBJECT_METHODS(name)								\
	((name ## __Methods *) DF__Object__get_methods(name ## __get_type()))
#define DF__OBJECT_INTERFACES(name)								\
	((name ## __Interfaces *) DF__Object__get_interfaces(name ## __get_type()))

#define DF__OBJECT_IS_A(object, name)								\
	(object==NULL ? FALSE : DF__Object__is_type(object->type, name ## __get_type()))

/* Only useful for downcasts, for upcasts use object->cast.<type> as that is statically typesafe
 * and this is only runtime typesafe */
#define DF__OBJECT_CAST(object, name)								\
	((name *)(object == NULL ? NULL :								\
		  (DF__Object__is_type_warn(object->type, name ## __get_type()) ?		\
		   object : NULL)))


/*
 * Definition of the DF__Object type
 */

#define DF__Object__CAST									\
	DF__Object *DF__Object;

#define DF__Object__METHODS									\
	void (* destroy)(DF__OBJECT_TYPE *self);

#define DF__Object__INTERFACES									\
	gpointer DF__Object__dummy;

#define DF__Object__DATA									\
	gpointer DF__Object__dummy;

#define DF__OBJECT_TYPE DF__Object
DF__OBJECT_DECLARATIONS(DF__Object,DF__Object)
#undef DF__OBJECT_TYPE DF__Object

DF__Object *	DF__Object__create		(guint type);
/* This will also emit a warning if type cannot be cast to istype */
gboolean	DF__Object__is_type		(guint type, guint istype);
/* This will also emit a warning if type cannot be cast to istype */
gboolean	DF__Object__is_type_warn	(guint type, guint istype);

typedef void	(*DF__Object__init)		(gpointer data);
typedef void	(*DF__Object__init_data)	(DF__Object *self);

gpointer	DF__Object__get_methods		(guint type);
gpointer	DF__Object__get_interfaces	(guint type);

guint 		DF__Object__init_type		(const char			*name,
						 guint				 parent_type,
						 DF__Object__init		 init_methods,
						 DF__Object__init		 init_interfaces,
						 DF__Object__init_data		 init_data,
						 guint				 sizeof_object,
						 guint				 sizeof_methods,
						 guint				 sizeof_interfaces);



#endif /* DF_OBJECT_H */
